unit Stres;
(***********************************************************************
 *                                                                     *
 *  Copyright (C) 1997-19..                                            *
 *  Arbeitsgruppe KI-Labor Fachbereich Informatik Universitaet Bremen  *
 *                                                                     *
 *  Version     :       2.0                                            *
 *                                                                     *
 *  Beschreibung:                                                      *
 *                                                                     *
 *  Autor       :       Erol S. Uzuner                                 *
 *                                                                     *
 *  Datum       :       Dezember 1997                                  *
 *                                                                     *
 *                                                                     *
 *  Aenderungen :                                                      *
 *  Datum        Autor          Beschreibung                           *
 *                                                                     *
 ***********************************************************************)

interface

uses Classes, SysUtils;

type
   NoSuchElementException = class(Exception);

   (* String Tokenizer Class
    * gets Parts of a string seperated by delimiters
    *)
   TStrTokenizer = class
       currentPosition, maxPosition : Integer;
       str, delimiters : String;
       retTokens: boolean;
       constructor Create(Str: String; delim : String; returnTokens: boolean);
       function hasMoreTokens: boolean;
       function hasMoreElements: boolean;
       function nextToken: String;
       function countTokens: Integer;
       procedure Reset(Str: String; delim : String; returnTokens: boolean);
       function isEmpty: boolean;
     private
       procedure skipDelimiters;
   end;


   function IsEmptyStr(const Str : string): Boolean;
   (* Replaces repStr with newStr *)
   function StrReplace(const str : string; const  repStr: String; const  newStr: String): String;
   (* Returns left string from count *)
   function LeftStr(const str: String; count: Integer): String;
   function RFillStr(const str: String; count: Integer; Char2Fill: char): String;
   function LFillStr(const str: String; count: Integer; Char2Fill: char): String;
   function BoolToStr(value : boolean): String;
   function StrToBool(const str: String): boolean;
   function RCutChars(chr: char; const str: String): String;
   function LCutChars(chr: char; const str: String): String;
   function TrimChars(ch : char; const str : String) : String;
   function StrBeginsWith(const subStr : String; const Str : string) : boolean;

//   function QPos(const SubStr, Str : String; Start : Cardinal = 1) : Cardinal;

   function CountEmptyStr(sarr : Array of string) : Integer;

implementation

 (*
  * Constructs a StrTokenizer on the specified String, using the
  * specified delimiter set.
  * @param str	        the input String
  * @param delim        the delimiter String
  * @param returnTokens returns delimiters as tokens or skip them
  *)
constructor TStrTokenizer.Create(Str : String; delim : String; returnTokens: boolean);
begin
   currentPosition := 1;
   self.str := str;
   maxPosition := Length(str) + 1;
   delimiters := delim;
   retTokens := returnTokens;
end;

function TStrTokenizer.isEmpty: boolean;
begin
   Result := str = '';
end;

(*
  * Resets the StrTokenizer on the specified String, using the
  * specified delimiter set.
  * @param str	        the input String
  * @param delim        the delimiter String
  * @param returnTokens returns delimiters as tokens or skip them
  *)
procedure TStrTokenizer.Reset(Str : String; delim : String; returnTokens: boolean);
begin
   currentPosition := 1;
   self.str := str;
   maxPosition := Length(str) + 1;
   delimiters := delim;
   retTokens := returnTokens;
end;

(*
 * Skips delimiters.
 *)
procedure TStrTokenizer.skipDelimiters;
begin
   if (NoT(retTokens) AND
       (currentPosition < maxPosition) AND
       (Pos(delimiters, str[currentPosition]) <> 0)) then
	    INC(currentPosition);
  
   (*	while (NoT(retTokens) AND
	       (currentPosition < maxPosition) AND
	       (Pos(delimiters, str[currentPosition]) <> 0)) do
	    INC(currentPosition);
    *)
end;

(*
 * Returns true if more tokens exist.
 *)
function TStrTokenizer.hasMoreTokens: boolean;
begin
	skipDelimiters;
	Result :=  (currentPosition < maxPosition);
end;

(*
 * Returns the next token of the String.
 * @exception NoSuchElementException If there are no more
 * tokens in the String.
 *)
 function TStrTokenizer.nextToken: String;
 var
    start : Integer;
 begin
        Result := '';

        if (currentPosition >= maxPosition) then
	    raise NoSuchElementException.Create('No Such Elements')
        else
        begin

           start := currentPosition;
          (* if ((currentPosition < maxPosition) AND
                  (Pos(delimiters, str[currentPosition]) = 0)) then
               INC(currentPosition);
          *) while ((currentPosition < maxPosition) AND
                  (Pos(delimiters, str[currentPosition]) = 0)) do
               INC(currentPosition);

           if (retTokens AND (start = currentPosition) AND
               (Pos(delimiters, str[currentPosition]) <> 0)) then
               INC(currentPosition);

           Result := copy(str, start, currentPosition - start);
        end;
      	skipDelimiters;
 end;

 (**
  * Returns true if the Enumeration has more elements.
  *)
  function TStrTokenizer.hasMoreElements: boolean;
  begin
	Result := hasMoreTokens;
  end;

(**
 * Returns the next number of tokens in the String using
 * the current deliminter set.  This is the number of times
 * nextToken() can return before it will generate an exception.
 * Use of this routine to count the number of tokens is faster
 * than repeatedly calling nextToken() because the substrings
 * are not constructed and returned for each token.
 *)
 function TStrTokenizer.countTokens: Integer;
 var
     count, currpos (*, start *): Integer;
 begin
	count := 0;
	currpos := currentPosition;

	while (currpos < maxPosition) do
        begin
	    (*
	     * This is just skipDelimiters(); but it does not affect
	     * currentPosition.
	     *)
	    while (NoT(retTokens) AND
		   (currpos < maxPosition) AND
		   (Pos(delimiters, str[currpos]) <> 0)) do
		INC(currpos);

	    if (currpos >= maxPosition) then
		break;

	   // start := currpos;

            if ((currpos < maxPosition) AND
		   (Pos(delimiters, str[currpos]) = 0)) then
		INC(currpos);

	    while ((currpos < maxPosition) AND
		   (Pos(delimiters, str[currpos]) = 0)) do
		INC(currpos);

	   (* if (retTokens AND (start = currpos) AND
		(Pos(delimiters, str[currpos]) <> 0)) then
		INC(currpos);
            *)
	    INC(count);

	end;
	Result := count;
end;

(* Replaces repStr with newStr in str *)
function StrReplace(const str : string; const repStr: String; const newStr: String): String;
var
   tstr : string;
   indx : Integer;
begin
     tstr := str;
     indx := Pos(repStr, tstr);

     Result := '';

     while indx > 0 do
     begin
          if indx > 1 then
               Result := Result + Copy(tstr, 1, indx-1) + newStr
          else
                Result := Result + newStr;

          tstr := Copy(tstr, indx+Length(repStr), Length(tstr));
          indx := Pos(repStr, tstr);
     end;
     Result := Result + tstr;
end;

(* Returns cuts the string at position count *)
function LeftStr(const str: String; count: Integer): String;
begin
     if count < 1 then Result := ''
     else if count > Length(str) then Result := str
     else Result := Copy(str, 1, count);
end;

function doFilling(Char2Fill:Char; count: Integer): string;
var
  i : integer;
begin
    Result := '';
    for i:=1 to count do
           Result := Result + Char2Fill;
end;

function RFillStr(const str: String; count: Integer; Char2Fill: char): String;
begin
     if count <= Length(str) then  Result := str
     else Result := str + doFilling(Char2Fill, count - Length(str));
end;

function LFillStr(const str : String; count : Integer; Char2Fill: char):String;
begin
   if count <= Length(str) then Result := str
   else Result := doFilling(Char2Fill, count - Length(str)) + str;
end;

function BoolToStr(value: boolean): String;
begin
   if value then Result := 'true'
   else Result := 'false';
end;

function StrToBool(const str: String): boolean;
begin
   if str = 'true' then Result := true
   else Result := false;
end;

(***********************************************************************
 *                                                                     *
 *  Name        :  IsEmptyString                                       *
 *                                                                     *
 *  Beschreibung:  Der bergebene String wird geprft ob er  Leer ist  *
 *                 oder nicht.                                         *
 *                                                                     *
 *  Rueckgabe   :  True : Wenn String leer oder nur aus SPACE'es       *
 *                        bestehend.                                   *
 *                 False : Wenn Sring nicht leer                       *
 *                                                                     *
 ***********************************************************************)
(*
function _IsEmptyStr(const Str : string): Boolean;
var
   cnt : Integer;
begin
  Result := TRUE;
  {Ist die Lnge unleich 0, knnte es angehen das der String
   aus SPACE'es besteht. deshalb berprfen wir dies auch.}
  if Length(Str) <> 0 then  begin
     for cnt := 1 to Length(Str) do begin
       if Str[cnt] > ' ' then begin
           Result := FALSE;
           break;
       end;
     end;
  end;
end;
  *)
function isEmptyStr(const str : String) : boolean;
asm
 // eax stores the param str
  test eax, eax;        // Nullpointer  or '' String
  jz @TRUE_END;          // go outta here
  xor ecx, ecx;         // position counter,
 @LOOP_1:
  cmp [eax-$04], ecx;   // we are done when Length(str) = position counter
  jz @TRUE_END;          // jump outta

{ --- Here we only checked tab and space ---
  mov dl, [eax+ ecx];   // current char
  cmp dl , $09;  // if current char = $09 ($09 = Ascii Char 9 = <TAB>)
  jz @DO_LOOP;          // we found a char which isnt a space :)
  cmp dl, $20;  // if current char <> $20 ($20 = Ascii Char 32 = <Space>)
  jnz @TRUE_END;        // we found a char which isnt a space :)
 @DO_LOOP:
}
  cmp byte ptr [eax+ ecx], $20; // every char less than 32 <SPACE> isn't regular
  jg @FALSE_END;        // we found a char regular char :)

  inc ecx;              // inc processed Chars
  jmp @LOOP_1;          // loop again with next char

 @TRUE_END:
   mov Result, true;     // Result is true we found no regular chars
   jmp @THE_END
 @FALSE_END:
   mov Result, false;   // this string isnt empty
 @THE_END:
end;


(* Sucht von rechts nach links nach einem Zeichen das
 * verschieden ist von ch, und schneidet dann den hinteren
 * teil ab
 *)
(*
function RCutChars(ch: char; const str: String): String;
var
  i : Word;
begin
    if Length(str) = 0 then Result := ''
    else begin
       i := Length(str);
       while i > 1 do begin
             if str[i] <> ch then break;
             DEC(i);
       end;
       if (i = 1) AND (str[1] = ch) then Result := ''
       else   Result := Copy(str, 1, i);
    end;
end;
*)
function RCutChars(chr: char; const str : String) : String;
asm
  { al beinhaltet den char
    edx den String
    ecx den result, wir benutzen aber ecx auch als Zaehler, deshalb speichern in ebx}
  push ebx; {register retten;}

  { Testen wir ob wir leeren String haben }
  test edx, edx;   { null String ?}
  jz @@EMPTY_STRING; { ja -> Wir haben nicht zu tun also leeren String }

 { Test auf null char entfllt weil sich das von alleine lst}
 //  test al, al;   // null char, ganzen String zurcksenden
 //  jz @COMPLETE_STRING;

  mov ebx, ecx; { Die Adresse des Ergebnisses behalten wir in ebx }
  { Nun schauen wir uns den String genauer an}
  mov ecx, [edx-$04]; { Lnge des Strings }
  { ecx beinhaltet nun den Positionszeiger
    edx + ecx ist die adresse des aktuellen Chars,
    zu beachten das edx + 0 den ersten char liefert also str[1]
    deshalb decrementieren wir edx  um mit dem Zaehler in ecx
    besser umzugehen.}
  dec edx;
 @@THE_LOOP:
  cmp byte ptr [edx+ecx], al; { vergleiche aktuellen char mit chr }
  jnz @@CUT_IT;                { wenn nicht gleich dann abschneiden }
  loop @@THE_LOOP;

 @@EMPTY_STRING:
  mov eax, ecx;  {ecx = result, Vorsicht hier}
  call System.@LStrClr;
  jmp @@THE_END;

 @@CUT_IT:
  inc  edx;           { String Zeiger auf die richtige position setzen }
  cmp ecx, [edx-$04]; { men wir denn was abschneiden }
  je @@COMPLETE_STRING;
{ LStrCopy parameter }
{     ->EAX     Source string                   }
{       EDX     index                           }
{       ECX     count                           }
{       [ESP+4] Pointer to result string        }
  push ebx;       { Result auf den Stack, durch aufruf von LStrCopy wieder ge-popt }
  mov eax, edx;   { String  Param 1}
  mov edx, 1;     { Index   Param 2}
  //mov ecx, ecx; { count   Param 3}
  call System.@LStrCopy;   { System.Copy(S; Index, Count: Integer): string; }
  jmp @@THE_END;

 @@COMPLETE_STRING:
  { Man kann nicht einfach die Adresse des Strings dem Result zuweisen,
    Delphi verwaltet intern Verweise auf Strings, wodurch wir jegliche
    Referenzen anmelden men. }
{ LStrLAsg parameter }
{ ->    EAX     pointer to dest }
{       EDX     source          }
  mov eax, ebx;   { Parameter 1 : Resultadresse in eax   }
                  { Parameter 2 : edx beinhaltet den Zeiger auf den String. siehe oben }
  call System.@LStrLAsg;

 @@THE_END:
  pop ebx;  { register herstellen }
end;


(* Sucht von links nach rechts nach einem Zeichen das
 * verschieden ist von ch, und schneidet dann den vorderen
 * teil ab
 *)
(*
function LCutChars(ch: char; const str: String): String;
var
    i : Word;
begin
    if Length(str) = 0 then Result := ''
    else begin
       i := 1;
       while i < Length(str) do
       begin
             if str[i] <> ch then break;
             INC(i);
       end;

       if (i = Length(str)) AND (str[i] = ch) then Result := ''
       else   Result := Copy(str, i, Length(str));
    end;
end;
*)
function LCutChars (chr: char; const str : String) : String;
asm
  { al beinhaltet den char
    edx den String
    ecx den result, wir benutzen aber ecx auch als Zaehler, deshalb speichern in ebx}
  push ebx; {register retten;}
  { Testen wir ob wir leeren String haben }
  test edx, edx;   { null String ?}
  jz @@EMPTY_STRING; { ja -> Wir haben nicht zu tun also ganz raus hier }

 { Test auf null char entfllt weil sich das von alleine lst}
 //  test al, al;   // null char, ganzen String zurcksenden
 //  jz @COMPLETE_STRING;

  mov ebx, ecx; { Die Adresse des Ergebnisses behalten wir in ebx }
  { Nun schauen wir uns den String genauer an}
  xor ecx, ecx; { Anfangsposition des Vergleichs }
  { ecx beinhaltet nun den Positionszeiger
    edx + ecx ist die adresse des aktuellen Chars,
    zu beachten das edx + 0 den ersten char liefert also str[1]}
 @@THE_LOOP:
  cmp byte ptr [edx+ecx], al; { vergleiche aktuellen char mit chr }
  jnz @@CUT_IT;                { wenn nicht gleich dann abschneiden }
  inc ecx;
  cmp ecx, [edx-$04];         { wenn wir am Ende angekommen sind leeren String liefern}
  jz @@EMPTY_STRING;
  jmp @@THE_LOOP;

 @@EMPTY_STRING:
  mov eax, ecx;  {ecx = result, Vorsicht hier}
  call System.@LStrClr;
  jmp @@THE_END;

 @@CUT_IT:
  test ecx, ecx;        { men wir denn was abschneiden }
  jz @@COMPLETE_STRING;
{ String Copy  parameter }
{     ->EAX     Source string                   }
{       EDX     index                           }
{       ECX     count                           }
{       [ESP+4] Pointer to result string        }
  push ebx;             { Result auf den Stack, durch aufruf von LStrCopy wieder ge-popt }
  mov eax, edx;         { String  Param 1}
  inc ecx;
  mov edx, ecx;         { Index   Param 2}
  mov ecx, [eax - $04]; { count   Param 3}
  call System.@LStrCopy;{ System.Copy(S; Index, Count: Integer): string; }
  jmp @@THE_END;

 @@COMPLETE_STRING:
  { Man kann nicht einfach die Adresse des Strings dem Result zuweisen,
    Delphi verwaltet intern Verweise auf Strings, wodurch wir jegliche
    Referenzen anmelden men }
{ LStrLAsg parameter }
{ ->    EAX     pointer to dest }
{       EDX     source          }
  mov eax, ebx;   { Parameter 1 : Resultadresse in eax   }
                  { Parameter 2 : edx beinhaltet den Zeiger auf den String }
  call System.@LStrLAsg;


 @@THE_END:
  pop ebx;  { register herstellen }
end;


function TrimChars(ch : char; const str : String) : String;
begin Result := LCutChars(ch, RCutChars(ch, str));
end;

(*****
 * Zaehlt die Leeren Strings und liefert die Anzahl zurck
 *)
function CountEmptyStr(sarr : Array of string) : Integer;
var i : Integer;
begin
  Result := 0;
  for i := 0 to High(sarr) do if IsEmptyStr(sarr[i]) then INC(Result);
end;

function StrBeginsWith(const subStr : String; const Str : string) : boolean;
begin Result := Pos(subStr, Str) = 1;
end;

end.
